# -*- coding: utf-8 -*-
r"""
Frequency-dependent foreground components.

This module implements the frequency-dependent component of common foreground
contaminants.

This package draws inspiration from FGBuster (Davide Poletti and Josquin Errard)
and BeFoRe (David Alonso and Ben Thorne).
"""

import inspect
import types
import numpy as np
from scipy import constants
from .model import Model


T_CMB = 2.72548
H_OVER_KT_CMB = constants.h * 1e9 / constants.k / T_CMB


def _flux2cmb(nu):
    """Converts flux to thermodynamics units"""
    x = H_OVER_KT_CMB * nu
    g2_min1 = (
        2.0
        * constants.k**3
        * T_CMB**2
        * x**4
        * np.exp(x)
        / (constants.h * constants.c * np.expm1(x)) ** 2
    )
    return 1.0 / g2_min1


def _rj2cmb(nu):
    x = H_OVER_KT_CMB * nu
    return (np.expm1(x) / x) ** 2 / np.exp(x)


def _bandpass_integration():
    """Bandpass integrated version of the caller

    The caller should have

        if isinstance(nu, list):
            return _bandpass_integration()

    at the very beginning.
    This function

    * iterates over the ``nu`` argument of the caller
      (while keeping all the other arguments fixed)
    * splits each element of the iteration in ``nu_band, transmittance``
    * integrates the caller function over the bandpass.
      ``np.trapz(caller(nu_band) * transmittance, nu_band)``
      Note that no normalization nor unit conversion is done to the
      transmittance
    * stacks the output of the iteration (the frequency dimension is the last)
      and returns it

    """
    # This piece of code is fairly complicated, we did because:
    # 1) We want to call eval on each element of the nu list (i.e. we iterate
    #    over the bandpasses) but we don't want to define a new eval_bandpass
    #    function for every class
    # 2) We don't want to use a decorator because it breaks the signature
    #    handling of eval and the modification of its defaults.
    #    _bandpass_integration does from the insides of eval the same thing that
    #    a decorator would do from the outside. This is achieved through the
    #    following pretty ugly kludge
    # Simpler code that achieve the same result is welcome

    # You are here because this function was called inside eval before any other
    # variable was defined.
    # We now retrieve the keyword arguments that were passed to eval because we
    # have to use them for the evaluation of eval on each bandpass
    # It assumes that _bandpass_integration was called inside
    # f(self, **kw) -- f is typically the eval method.
    frame = inspect.currentframe().f_back
    kw = frame.f_locals
    self = kw["self"]
    del kw["self"]  # self was in the locals but is not a keyword argument

    # We create a copy of eval itself, we'll call it for each bandpass
    f = types.FunctionType(frame.f_code, frame.f_globals)
    # Store the nu-transmittance list because the nu keyword argumnt has to be
    # modified with the frequencies of each bandpass
    nus_transmittances = kw["nu"]

    # Get the shape of the output from the result of the first bandpass
    kw["nu"] = nus_transmittances[0][0]
    res = np.trapz(f(self, **kw) * nus_transmittances[0][1], kw["nu"])
    # Append the frequency dimension and put res in its first entry
    res = res[..., np.newaxis] * np.array([1.0] + [0.0] * (len(nus_transmittances) - 1))

    # Fill the remaining entries by iterating over the rest of the bandpasses
    for i_band, (nu, transmittance) in enumerate(nus_transmittances[1:], 1):
        kw["nu"] = nu
        res[..., i_band] = np.trapz(f(self, **kw) * transmittance, nu)

    return res


class PowerLaw(Model):
    r"""Power Law

    .. math:: f(\nu) = (\nu / \nu_0)^{\beta}
    """

    def eval(self, nu=None, beta=None, nu_0=None):
        """Evaluation of the SED

        Parameters
        ----------
        nu: float or array
            Frequency in the same units as `nu_0`. If array, the shape is
            ``(freq)``.
        beta: float or array
            Spectral index. If array, the shape is ``(...)``.
        nu_0: float or array
            Reference frequency in the same units as `nu`. If array, the shape
            is ``(...)``.

        Returns
        -------
        sed: ndarray
            If `nu` is an array, the shape is ``(..., freq)``.
            If `nu` is scalar, the shape is ``(..., 1)``.
            Note that the last dimension is guaranteed to be the frequency.

        Note
        ----
        The extra dimensions ``...`` in the output are the broadcast of the
        ``...`` in the input (which are required to be broadcast-compatible).

        Examples
        --------

        - T, E and B synchrotron SEDs with the same reference frequency but
          different spectral indices. `beta` is an array with shape ``(3)``,
          `nu_0` is a scalar.

        - SEDs of synchrotron and dust (approximated as power law). Both `beta`
          and `nu_0` are arrays with shape ``(2)``

        """
        if isinstance(nu, list):
            return _bandpass_integration()

        beta = np.array(beta)[..., np.newaxis]
        nu_0 = np.array(nu_0)[..., np.newaxis]
        return (nu / nu_0) ** beta * (_rj2cmb(nu) / _rj2cmb(nu_0))


class Synchrotron(PowerLaw):
    """Alias of :class:`PowerLaw`"""

    pass


class ModifiedBlackBody(Model):
    r"""Modified black body in K_RJ

    .. math:: f(\nu) = (\nu / \nu_0)^{\beta + 1} / (e^x - 1)

    where :math:`x = h \nu / k_B T_d`
    """

    def eval(self, nu=None, nu_0=None, temp=None, beta=None):
        """Evaluation of the SED

        Parameters
        ----------
        nu: float or array
            Frequency in GHz.
        beta: float or array
            Spectral index.
        temp: float or array
            Dust temperature.
        nu_0: float
            Reference frequency in Hz.

        Returns
        -------
        sed: ndarray
            The last dimension is the frequency dependence.
            The leading dimensions are the broadcast between the hypothetic
            dimensions of `beta` and `temp`.
        """
        if isinstance(nu, list):
            return _bandpass_integration()

        beta = np.array(beta)[..., np.newaxis]
        temp = np.array(temp)[..., np.newaxis]
        x = 1e9 * constants.h * nu / (constants.k * temp)
        x_0 = 1e9 * constants.h * nu_0 / (constants.k * temp)
        res = (nu / nu_0) ** (beta + 1.0) * np.expm1(x_0) / np.expm1(x)
        return res * (_rj2cmb(nu) / _rj2cmb(nu_0))


class CIB(ModifiedBlackBody):
    """Alias of :class:`ModifiedBlackBOdy`"""

    pass


class ThermalSZ(Model):
    r"""Thermal Sunyaev-Zel'dovich in K_CMB

    This class implements the

    .. math:: f(\nu) = x \coth(x/2) - 4

    where :math:`x = h \nu / k_B T_CMB`
    """

    @staticmethod
    def f(nu):
        x = constants.h * (nu * 1e9) / (constants.k * T_CMB)
        return x / np.tanh(x / 2.0) - 4.0

    def eval(self, nu=None, nu_0=None):
        """Compute the SED with the given frequency and parameters.

        nu : float
            Frequency in GHz.
        T_CMB (optional) : float
        """
        if isinstance(nu, list):
            return _bandpass_integration()

        return ThermalSZ.f(nu) / ThermalSZ.f(nu_0)


class FreeFree(Model):
    r"""Free-free

    .. math:: f(\nu) = EM * ( 1 + log( 1 + (\nu_{ff} / \nu)^{3/\pi} ) )
    .. math:: \nu_{ff} = 255.33e9 * (Te / 1000)^{3/2}
    """

    def eval(self, nu=None, EM=None, Te=None):
        """Evaluation of the SED

        Parameters
        ----------
        nu: float or array
            Frequency in the same units as `nu_0`. If array, the shape is
            ``(freq)``.
        EM: float or array
            Emission measure in cm^-6 pc (usually around 300). If array, the shape is ``(...)``.
        Te: float or array
            Electron temperature (typically around 7000). If array, the shape is ``(...)``.

        Returns
        -------
        sed: ndarray
            If `nu` is an array, the shape is ``(..., freq)``.
            If `nu` is scalar, the shape is ``(..., 1)``.
            Note that the last dimension is guaranteed to be the frequency.

        Note
        ----
        The extra dimensions ``...`` in the output are the broadcast of the
        ``...`` in the input (which are required to be broadcast-compatible).

        Examples
        --------

        - Free-free emission in temperature.

        """
        if isinstance(nu, list):
            return _bandpass_integration()

        EM = np.array(EM)[..., np.newaxis]
        Te = np.array(Te)[..., np.newaxis]
        Teff = (Te / 1.0e3) ** (1.5)
        nuff = 255.33e9 * Teff
        gff = 1.0 + np.log(1.0 + (nuff / nu) ** (np.sqrt(3) / np.pi))
        print("warning: I need to check the units on this")
        return EM * gff


class ConstantSED(Model):
    """Frequency-independent component."""

    def eval(self, nu=None, amp=1.0):
        """Evaluation of the SED

        Parameters
        ----------
        nu: float or array
            It just determines the shape of the output.
        amp: float or array
            Amplitude (or set of amplitudes) of the constant SED.

        Returns
        -------
        sed: ndarray
            If `nu` is an array, the shape is ``amp.shape + (freq)``.
            If `nu` is scalar, the shape is ``amp.shape + (1)``.
            Note that the last dimension is guaranteed to be the frequency.
        """
        if isinstance(nu, list):
            return _bandpass_integration()

        amp = np.array(amp)[..., np.newaxis]
        return amp * np.ones_like(np.array(nu))


class FreeSED(Model):
    """Frequency-dependent component for which every entries of the SED are specifified."""

    def eval(self, nu=None, sed=None):
        """Evaluation of the SED

        Parameters
        ----------
        nu: float or array
            It just determines the shape of the output.
        sed: float or array
            Values of the SED. Must be the same shape as nu.
            There is no normalisation, entries are used as is.

        Returns
        -------
        sed: ndarray
            If `nu` is an array, the shape is ``amp.shape + (freq)``.
            If `nu` is scalar, the shape is ``amp.shape + (1)``.
            Note that the last dimension is guaranteed to be the frequency.
        """
        if isinstance(nu, (list, np.ndarray)):
            try:
                assert len(nu) == len(sed[..., :])
            except:
                print("SED and nu must have the same shape.")
        if isinstance(nu, list):
            return _bandpass_integration()
        return np.asarray(sed)


class Join(Model):
    """Join several SED models together"""

    def __init__(self, *seds, **kwargs):
        """Join several SED models together

        Parameters
        ----------
        *sed:
            Sequence of SED models to be joined together
        """
        self._seds = seds
        self.set_defaults(**kwargs)

    def set_defaults(self, **kwargs):
        if "kwseq" in kwargs:
            for sed, sed_kwargs in zip(self._seds, kwargs["kwseq"]):
                sed.set_defaults(**sed_kwargs)

    def _get_repr(self):
        return {type(self).__name__: [sed._get_repr() for sed in self._seds]}

    @property
    def defaults(self):
        return {"kwseq": [sed.defaults for sed in self._seds]}

    def eval(self, kwseq=None):
        """Compute the SED with the given frequency and parameters.

        *kwseq
            The length of ``kwseq`` has to be equal to the number of SEDs
            joined. ``kwseq[i]`` is a dictionary containing the keyword
            arguments of the ``i``-th SED.
        """
        if kwseq:
            seds = [sed(**kwargs) for sed, kwargs in zip(self._seds, kwseq)]
        else:  # Handles the case in which no parameter has to be passed
            seds = [sed() for sed in self._seds]
        res = np.empty((len(seds),) + np.broadcast(*seds).shape)
        for i in range(len(seds)):
            res[i] = seds[i]
        return res



r"""
Models of cross-spectra

This module draws inspiration from FGBuster (Davide Poletti and Josquin Errard)
and BeFoRe (David Alonso and Ben Thorne).
"""

import numpy as np
from . import frequency as fgf
from . import power as fgp
from .model import Model


class Sum(Model):
    """Sum the cross-spectra of uncorrelated components"""

    def __init__(self, *crosses, **kwargs):
        """

        Parameters
        ----------
        *sed:
            Sequence of SED models to be joined together
        """
        self._crosses = crosses
        self.set_defaults(**kwargs)

    def set_defaults(self, **kwargs):
        if "kwseq" in kwargs:
            for cross, cross_kwargs in zip(self._crosses, kwargs["kwseq"]):
                cross.set_defaults(**cross_kwargs)

    def _get_repr(self):
        return {type(self).__name__: [cross._get_repr() for cross in self._crosses]}

    @property
    def defaults(self):
        return {"kwseq": [cross.defaults for cross in self._crosses]}

    def eval(self, kwseq=None):
        """Compute the sum of the cross-spectra

        *kwseq:
            The length of ``kwseq`` has to be equal to the number of
            cross-spectra summed. ``kwseq[i]`` is a dictionary containing the
            keyword arguments of the ``i``-th cross-spectrum.
        """
        if kwseq:
            crosses = (cross(**kwargs) for cross, kwargs in zip(self._crosses, kwseq))
        else:  # Handles the case in which no parameter has to be passed
            crosses = (cross() for cross in self._crosses)

        res = next(crosses)
        for cross_res in crosses:
            res = res + cross_res  # Unnecessary copies can be avoided

        return res

    def eval_terms(self, kwseq=None):
        """Compute the sum of the cross-spectra

        *kwseq:
            The length of ``kwseq`` has to be equal to the number of
            cross-spectra summed. ``kwseq[i]`` is a dictionary containing the
            keyword arguments of the ``i``-th cross-spectrum.
        """
        if kwseq:
            return [cross(**kwargs) for cross, kwargs in zip(self._crosses, kwseq)]
        return []


class FactorizedCrossSpectrum(Model):
    r"""Factorized cross-spectrum

    Cross-spectrum of **one** component for which the scaling in frequency
    and in multipoles are factorizable

    .. math:: xC_{\ell}^{(ij)} = f(\nu_j) f(\nu_i) C_{\ell}

    Parameters
    ----------
    sed : callable
        :math:`f(\nu)`. It returns an array with shape ``(..., freq)``.
        It can be :class:`fgspectra.frequency.SED`.
    cl : callable
        :math:`C_\ell`. It returns an array with shape ``(..., ell)``.
        It can be :class:`fgspectra.power.PowerSpectrum`

    Note
    ----
    The two (optional) sets of extra dimensions ``...`` must be
    broadcast-compatible.
    """

    def __init__(self, sed, cl, **kwargs):
        self._sed = sed
        self._cl = cl
        self.set_defaults(**kwargs)

    def set_defaults(self, **kwargs):
        if "sed_kwargs" in kwargs:
            self._sed.set_defaults(**kwargs["sed_kwargs"])
        if "cl_kwargs" in kwargs:
            self._cl.set_defaults(**kwargs["cl_kwargs"])

    @property
    def defaults(self):
        return {"sed_kwargs": self._sed.defaults, "cl_kwargs": self._cl.defaults}

    def _get_repr(self):
        sed_repr = self._sed._get_repr()
        key = list(sed_repr.keys())[0]
        sed_repr[key + " (SED)"] = sed_repr.pop(key)

        cl_repr = self._cl._get_repr()
        key = list(cl_repr.keys())[0]
        cl_repr[key + " (Cl)"] = cl_repr.pop(key)

        return {type(self).__name__: [sed_repr, cl_repr]}

    def eval(self, sed_kwargs={}, cl_kwargs={}):
        """Compute the model at frequency and ell combinations.

        Parameters
        ----------
        sed_args : list
            Arguments for which the `sed` is evaluated.
        cl_args : list
            Arguments for which the `cl` is evaluated.

        Returns
        -------
        cross : ndarray
            Cross-spectrum. The shape is ``(..., freq, freq, ell)``.
        """
        f_nu = self._sed(**sed_kwargs)
        cl = self._cl(**cl_kwargs)
        if f_nu.shape[0] != cl.shape[-1] or (f_nu.shape[0] == 1 and cl.shape[-1] == 1):
            f_nu = f_nu[np.newaxis]

        return np.einsum("l...i,l...j,...l->...ijl", f_nu, f_nu, cl)


class DecorrelatedFactorizedCrossSpectrum(Model):
    r""" Decorrelated factorized cross spectrum

    Cross-spectrum of **one** component for which the scaling in frequency
    and in multipoles are factorizable. The scaling in frequency is not exact 
    and some decorrelation is specified.

    .. math:: xC_{\ell}^{(ij)} = f_{decor}(\nu_i, \nu_j) f(\nu_j) f(\nu_i) C_{\ell}
    
    This model implements a decorrelation by rigidely rescalling a decorelation ratio 
    with respect to a reference cross-term. For example, for 3 frequencies, with the
    reference taken at $\nu_1\times\nu_3$: 
         
    .. math::  f_{decor} = I(3) + \Delta_{decor} \begin{pmatrix} 
        0 & r_{12} & 1 \\         
        r_{12} & 0 & r_{23} \\              
        1 & r_{23} & 0 
        \end{pmatrix} 

    Parameters
    ----------
    sed : callable
        :math:`f(\nu)`. It returns an array with shape ``(..., freq)``.
        It can be :class:`fgspectra.frequency.SED`.
    cl : callable
        :math:`C_\ell`. It returns an array with shape ``(..., ell)``.
        It can be :class:`fgspectra.power.PowerSpectrum`

    Note
    ----
    The two (optional) sets of extra dimensions ``...`` must be
    broadcast-compatible.
        
    """

    def __init__(self, sed, cl, **kwargs):
        self._sed = sed
        self._cl = cl
        self.set_defaults(**kwargs)

    def set_defaults(self, **kwargs):
        if "sed_kwargs" in kwargs:
            self._sed.set_defaults(**kwargs["sed_kwargs"])
        if "cl_kwargs" in kwargs:
            self._cl.set_defaults(**kwargs["cl_kwargs"])
        kw_decor = {
            key: kwargs[key]
            for key in kwargs.keys()
            if key not in ["sed_kwargs", "cl_kwargs"]
        }
        super().set_defaults(**kw_decor)

    @property
    def defaults(self):
        kw_decor = {
            key: super().defaults[key]
            for key in super().defaults.keys()
            if key not in ["kwargs"]
        }
        return {
            **kw_decor,
            "sed_kwargs": self._sed.defaults,
            "cl_kwargs": self._cl.defaults,
        }

    def _get_repr(self):
        decor_repr = super()._get_repr()
        key = list(decor_repr.keys())[0]
        decor_repr[key] = {
            k: decor_repr[key][k]
            for k in decor_repr[key].keys()
            if k not in ["sed_kwargs", "cl_kwargs"]
        }
        decor_repr["Decorrelation"] = decor_repr.pop(key)

        sed_repr = self._sed._get_repr()
        key = list(sed_repr.keys())[0]
        sed_repr[key + " (SED)"] = sed_repr.pop(key)

        cl_repr = self._cl._get_repr()
        key = list(cl_repr.keys())[0]
        cl_repr[key + " (Cl)"] = cl_repr.pop(key)

        return {type(self).__name__: [decor_repr, sed_repr, cl_repr]}

    def eval(self, decor=None, f_decor=None, sed_kwargs={}, cl_kwargs={}):
        """Compute the model at frequency and ell combinations.

        Parameters
        ----------
        sed_args : list
            Arguments for which the `sed` is evaluated.
        cl_args : list
            Arguments for which the `cl` is evaluated.
        decor : float
            Decorelation facot, by which f_decor is rescaled
        f_decor : ndarray
            Array of the decorelation scaling in frequency.
            Shape is ``(freq, freq)``.
        Returns
        -------
        cross : ndarray
            Cross-spectrum. The shape is ``(..., freq, freq, ell)``.
        """

        decorrelation = np.eye(f_decor.shape[0]) + decor * f_decor

        f_nu = self._sed(**sed_kwargs)
        cl = self._cl(**cl_kwargs)
        if f_nu.shape[0] != cl.shape[-1] or (f_nu.shape[0] == 1 and cl.shape[-1] == 1):
            f_nu = f_nu[np.newaxis]

        res = np.einsum("ij,l...i,l...j,...l->...ijl", decorrelation, f_nu, f_nu, cl)

        return res


class CorrelatedFactorizedCrossSpectrum(FactorizedCrossSpectrum):
    r"""Factorized cross-spectrum of correlated components

    Cross-spectrum of multiple correlated components for which the scaling in
    frequency and in multipoles are factorizable.

    .. math::

        xC_\ell^{\nu_i\ \nu_j} = \sum_{kn} f^k(\nu_j) f^n(\nu_i) C^{kn}_\ell

    where :math:`k` and :math:`n` are component indices.

    Parameters
    ----------
    sed : callable
        :math:`f(\nu)`. It returns an array with shape ``(comp, ..., freq)``.
        It can be :class:`fgspectra.frequency.SED`.
    cl_args : callable
        :math:`C_\ell`. It returns an array with shape
        ``(..., comp, comp, ell)``.
        It can, for example, any of the models in :class:`fgspectra.power`.

    Note
    ----
    The two (optional) sets of extra dimensions ``...`` must be
    broadcast-compatible.
    """

    def __init__(self, sed, cl, **kwargs):
        self._sed = sed
        self._cl = cl
        self.set_defaults(**kwargs)

    def eval(self, sed_kwargs={}, cl_kwargs={}):
        """Compute the model at frequency and ell combinations.

        Parameters
        ----------
        sed_args : list
            Arguments for which the `sed` is evaluated.
        cl_args : list
            Arguments for which the `cl` is evaluated.

        Returns
        -------
        cross : ndarray
            Cross-spectrum. The shape is ``(..., freq, freq, ell)``.
        """

        f_nu = self._sed(**sed_kwargs)
        cl = self._cl(**cl_kwargs)
        if f_nu.shape[0] != cl.shape[-1]:
            f_nu = f_nu[:, np.newaxis]

        return np.einsum("kl...i,nl...j,...knl->...ijl", f_nu, f_nu, cl)


class PowerLaw(FactorizedCrossSpectrum):
    """Single Power law in both fequency and multipoles

    See :class:`fgspectra.frequency.PowerLaw` for the frequency dependence and
    :class:`fgspectra.power.PowerLaw` for the ell-dependence.
    """

    def __init__(self, **kwargs):
        super().__init__(fgf.PowerLaw(), fgp.PowerLaw())
        self.set_defaults(**kwargs)


class CorrelatedPowerLaw(CorrelatedFactorizedCrossSpectrum):
    """Correlated Power law in both fequency and multipoles

    See :class:`fgspectra.frequency.PowerLaw` for the frequency dependence and
    :class:`fgspectra.power.PowerLaw` for the ell-dependence.
    """

    def __init__(self, **kwargs):
        super().__init__(fgf.PowerLaw(), fgp.CorrelatedPowerLaws())
        self.set_defaults(**kwargs)


class CorrelatedDustSynchrotron(CorrelatedFactorizedCrossSpectrum):
    """CorrelatedDustSynchrotron

    Correlated power law and modified black body, both with power law amplitude
    """

    def __init__(self, **kwargs):
        super().__init__(
            fgf.Join(fgf.ModifiedBlackBody(), fgf.PowerLaw()), fgp.CorrelatedPowerLaws()
        )
        self.set_defaults(**kwargs)


class SZxCIB(CorrelatedFactorizedCrossSpectrum):
    def __init__(self, **kwargs):
        sed = fgf.Join(fgf.ThermalSZ(), fgf.CIB())
        super().__init__(sed, fgp.SZxCIB_Addison2012())
        self.set_defaults(**kwargs)


class SZxCIB_Choi2020(CorrelatedFactorizedCrossSpectrum):
    def __init__(self, **kwargs):
        sed = fgf.Join(fgf.ThermalSZ(), fgf.CIB())
        power_spectra = [
            fgp.PowerSpectrumFromFile(fgp._get_power_file("tsz_150_bat")),
            fgp.PowerSpectrumFromFile(fgp._get_power_file("cib_Choi2020")),
            fgp.PowerSpectrumFromFile(fgp._get_power_file("sz_x_cib")),
        ]
        cl = fgp.PowerSpectraAndCovariance(*power_spectra)
        super().__init__(sed, cl)
        self.set_defaults(**kwargs)



from abc import ABC, abstractmethod
import types
import inspect
import yaml
import numpy as np
from copy import deepcopy


class Model(ABC):
    """Abstract class for model definition

    A model is a class that has one purpose: evaluating the model. What
    evaluation means is defined by the `eval` method that any child class has to
    override.

    If the `eval` method of a hypotetical ``Child`` class calls the `eval`
    method of other `Model`s, it is likely that ``Child`` has also to override
    `set_defaults`, `defaults` and `_get_repr`
    """

    def __init__(self, **kwargs):
        """You can set defaults at construction time"""
        self.set_defaults(**kwargs)

    def set_defaults(self, **kwargs):
        """Change the defaults of the evaluation

        Call ``set_defaults`` with the same arguments (or a subset) of ``eval``
        (or ``__call__``).
        This will change the default value of those arguments in ``eval``. If
        `kwargs` has keys that are not arguments of `eval`, these keys are
        ignored.

        Examples
        --------

        >>> class PrintABC(Model):
        ...     def eval(self, a=1, b=2, c=3):
        ...         print(a, b, c)
        ...
        >>> print_abc = PrintABC()
        >>> print_abc.eval()
        1 2 3
        >>> print_ab10c = PrintABC()
        >>> print_ab10c.set_defaults(b=10)
        >>> print_ab10c.eval()  # The default of b has changed
        1 10 3
        >>> print_abc.eval()  # Without affecting other instances
        1 2 3

        """
        # This is a dirty business, and has limitations. The most important
        # known one is that it breaks with decorated eval
        # To change the default values of the eval method
        # we have to bind a new method to the instance
        # (not the class, otherwise we change the defaults of all the instances)

        if not kwargs:  # No default to be replaced
            return

        # Make a deep copy of the eval function, assuming no variadic arguments
        func = self.eval.__func__

        new_func = types.FunctionType(
            func.__code__,
            func.__globals__,
            name=func.__name__,
            argdefs=func.__defaults__,
            closure=func.__closure__,
        )

        # Build the defaults from scratch, assuming no positional argument
        # Update the defaults that have a corresponding key in kwargs
        # Note that if kwargs has keys that are not in the arguments,
        # these keys are ignored
        fas = inspect.getfullargspec(new_func)
        defaults = []
        for arg, default in zip(fas.args[1:], fas.defaults):
            # [1:] skips self, the only positional argument
            if arg in kwargs:
                defaults.append(kwargs[arg])
            else:
                defaults.append(default)

        new_func.__defaults__ = tuple(defaults)

        # Bind the new method to the instance
        self.eval = types.MethodType(new_func, self)

    @property
    def defaults(self):
        """The current defaults

        The current defaults as you would pass them to the eval method.

        In particular, since all the arguments have to be keyword arguments, it
        is a nested set of dictionaries with a key for each argument of the
        `eval` methods in the model.

        >>> model.eval() == model.eval(**model_child.defaults)
        True

        """
        return {
            name: par.default
            for name, par in inspect.signature(self.eval).parameters.items()
        }

    def __call__(self, *args, **kwargs):
        # __call__ can't be rebound, that's why we use eval
        return self.eval(*args, **kwargs)

    @abstractmethod
    def eval(self, **kwargs):
        """Evaluation of the model

        Note
        ----
        The signature must contain

        - only keyword arguments (except for ``self``)
        - no variadic argument (*args, **kwargs)

        and

        - no decoration

        """
        pass

    def _get_repr(self):
        """Readable version of defaults

        In particular, it is also a nested set of dictionaries with a key for
        each argument of the `eval` methods in the model.

        If `eval` calls the `eval` of other classes, you typically want to
        call the _get_repr method of those classes and collect the result in
        some meaningful dictionary
        """
        res = {}
        for key, val in self.defaults.items():
            res[key] = val.tolist() if isinstance(val, np.ndarray) else val
        return {type(self).__name__: res}

    def __repr__(self):
        # Turns the nested set of dictionaries into a string to be printed
        return yaml.dump(self._get_repr())

    def _key_value_iteration(self, dict_list_tuple):
        if isinstance(dict_list_tuple, dict):
            return dict_list_tuple.items()
        if isinstance(dict_list_tuple, (list, tuple)):
            return enumerate(dict_list_tuple)
        raise TypeError("Only dict, list or tuple are allowed")

    def _update_path_nones(self):
        self._path_nones = []

        def search_for_none(kwargs, base=[]):
            for key, val in self._key_value_iteration(kwargs):
                if val is None:
                    self._path_nones.append(base + [key])
                else:
                    try:
                        search_for_none(val, base + [key])
                    except TypeError:
                        pass

        search_for_none(self.defaults)

    def prepare_for_arrays(self, template_kwargs):
        """Preapre for using arrays

        reference_kwargs are necessary to define the shape you expect for each
        argument, which are necessary to convert from array to nested
        dictionaries of arguments. They are required to have at least all the
        arguments for which the default is None.
        """
        self._template_kwargs = deepcopy(template_kwargs)
        self._update_path_nones()

    def kwargs2array(self, kwargs):
        """(Nested) dictionaries to float array"""
        if not hasattr(self, "_path_nones"):
            raise RuntimeError("You have to call prepare_for_arrays first")
        res_list = []
        for path in self._path_nones:
            val = kwargs
            for p in path:
                val = val[p]
            res_list.append(np.array(val).reshape(-1))

        return np.concatenate(res_list)

    def array2kwargs(self, x):
        if not hasattr(self, "_path_nones"):
            raise RuntimeError("You have to call prepare_for_arrays first")
        for path in self._path_nones:
            inner_path = path
            inner_kwargs = self._template_kwargs
            while len(inner_path) > 1:
                inner_kwargs = inner_kwargs[inner_path[0]]
                inner_path = inner_path[1:]

            ref_val = inner_kwargs[inner_path[0]]
            try:
                # It's an array
                inner_kwargs[inner_path[0]] = x[: ref_val.size].reshape(ref_val.shape)
                x = x[ref_val.size :]
            except AttributeError:
                # It's a float
                inner_kwargs[inner_path[0]] = x[0]
                x = x[1:]
        return self._template_kwargs

    def eval_array(self, x):
        return self.eval(**self.array2kwargs(x))



import fgspectra.frequency as fgf
from fgspectra.model import Model
import numpy as np
from numpy.testing import assert_allclose as aac


def test_bandpass_integration():
    frequency = np.array([[10.0, 20.0, 40.0], [60.0, 90.0, 100.0]])

    transmittance = np.array([[1.0, 2.0, 4.0], [1.0, 3.0, 1.0]])

    class Mock(Model):
        def eval(self, nu=None, par=None):
            if isinstance(nu, list):
                return fgf._bandpass_integration()
            return np.ones_like(nu) * np.array(par)[..., np.newaxis]

    mock = Mock()
    nus = list(zip(frequency, transmittance))

    ress = mock(nus, 1)
    refs = [75.0, 80.0]
    for res, ref in zip(refs, ress):
        aac(res, ref)

    ress = mock(nus, np.array([1.0, 2.0]))
    refs = np.array([[75.0, 80.0], [150.0, 160.0]])
    for res, ref in zip(refs, ress):
        aac(res, ref)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""fgspectra: GitHub <https://github.com/simonsobs/fgspectra>."""

from setuptools import setup

setup(
    name="fgspectra",
    version="1.1.0",
    description="Foreground SED and power spectrum library",
    author="Simons Observatory fgspectra crew",
    author_email="",
    packages=["fgspectra"],
    python_requires=">3.9",
    install_requires=[
        "scipy",
        "pyyaml",
    ],
    include_package_data=True,
)



r"""
Power spectrum

This module implements the ell-dependent component of common foreground
contaminants.

This module draws inspiration from FGBuster (Davide Poletti and Josquin Errard)
and BeFoRe (David Alonso and Ben Thorne).
"""

import os
import importlib.resources as ilr

import numpy as np
from .model import Model


def _get_power_file(model):
    """File path for the named model"""
    data_path = ilr.files("fgspectra") / "data"
    filename = os.path.join(data_path, "cl_%s.dat" % model)
    if os.path.exists(filename):
        return filename
    raise ValueError("No template for model " + model)


class PowerSpectrumFromFile(Model):
    """Power spectrum loaded from file(s)

    Parameters
    ----------
    filenames: array_like of strings
        File(s) to load. It can be a string or any (nested) sequence of strings

    Examples
    --------

    >>> ell = range(5)

    Power spectrum of a single file

    >>> my_file = 'cl.dat'
    >>> ps = PowerSpectrumFromFile(my_file)
    >>> ps(ell).shape
    (5)
    >>> ps = PowerSpectrumFromFile([my_file])  # List
    >>> ps(ell).shape
    (1, 5)

    Two correlated components

    >>> my_files = [['cl_comp1.dat', 'cl_comp1xcomp2.dat'],
    ...             ['cl_comp1xcomp2.dat', 'cl_comp2.dat']]
    >>> ps = PowerSpectrumFromFile(my_files)
    >>> ps(ell).shape
    (2, 2, 5)

    """

    def __init__(self, filenames, **kwargs):
        """

        The file format should be two columns, ell and the spectrum.
        """
        filenames = np.array(filenames)
        self._cl = np.empty(filenames.shape + (0,))

        for i, filename in np.ndenumerate(filenames):
            ell, spec = np.genfromtxt(filename, unpack=True)
            ell = ell.astype(int)
            # Make sure that the new spectrum fits self._cl
            n_missing_ells = ell.max() + 1 - self._cl.shape[-1]
            if n_missing_ells > 0:
                pad_width = [(0, 0)] * self._cl.ndim
                pad_width[-1] = (0, n_missing_ells)
                self._cl = np.pad(
                    self._cl, pad_width, mode="constant", constant_values=0
                )

            self._cl[i + (ell,)] = spec
        self.set_defaults(**kwargs)

    def eval(self, ell=None, ell_0=None, amp=1.0):
        """Compute the power spectrum with the given ell and parameters."""
        #### TEMPORARY ####
        if ell_0 is None:
            print("NO NORMALISATION")
            return amp * self._cl[..., ell]
        else:
            return amp * self._cl[..., ell] / self._cl[..., ell_0, np.newaxis]


class PowerLawExtendedTemplate(PowerSpectrumFromFile):
    """ " Power Spectrum from a file, extended by a power law on small scales."""

    def eval(self, ell=None, ell_0=None, alpha=None, amp=1.0):
        """Compute the power spectrum with the given ell and parameters."""
        # TODO : the switch from template to power law is at ell_0 for now !
        res = amp * self._cl[..., ell] / self._cl[..., ell_0, np.newaxis]
        res[..., ell > ell_0] = amp * (ell[ell > ell_0] / ell_0) ** alpha
        return res


class PowerLawRescaledTemplate(PowerSpectrumFromFile):
    """Power Spectrum from a file, rescaled by a power law."""

    def eval(self, ell=None, ell_0=None, alpha=None, amp=1.0):
        """Compute the power spectrum with the given ell and parameters."""
        return (
            amp
            * self._cl[..., ell]
            / self._cl[..., ell_0, np.newaxis]
            * (ell / ell_0) ** alpha
        )


class tSZ_150_bat(PowerSpectrumFromFile):
    """PowerSpectrum for Thermal Sunyaev-Zel'dovich (Dunkley et al. 2013)."""

    def __init__(self, **kwargs):
        """Intialize object with parameters."""
        super().__init__(_get_power_file("tsz_150_bat"), **kwargs)


class kSZ_bat_full(PowerSpectrumFromFile):
    """
    PowerSpectrum for Kinematic Sunyaev-Zel'dovich (Dunkley et al. 2013).
    Contains both contributions from late-time kSZ and reionization kSZ.
    """

    def __init__(self, **kwargs):
        """Intialize object with parameters."""
        super().__init__(_get_power_file("ksz_bat_full"), **kwargs)


class kSZ_bat(PowerSpectrumFromFile):
    """PowerSpectrum for Kinematic Sunyaev-Zel'dovich (Dunkley et al. 2013)."""

    def __init__(self, **kwargs):
        """Intialize object with parameters."""
        super().__init__(_get_power_file("ksz_bat"), **kwargs)


class PowerLaw(Model):
    r"""Power law

    .. math:: C_\ell = (\ell / \ell_0)^\alpha
    """

    def eval(self, ell=None, alpha=None, ell_0=None, amp=1.0):
        """

        Parameters
        ----------
        ell: float or array
            Multipole
        alpha: float or array
            Spectral index.
        ell_0: float
            Reference ell
        amp: float or array
            Amplitude, shape must be compatible with `alpha`.

        Returns
        -------
        cl: ndarray
            The last dimension is ell.
            The leading dimensions are the hypothetic dimensions of `alpha`
        """
        alpha = np.array(alpha)[..., np.newaxis]
        amp = np.array(amp)[..., np.newaxis]
        return amp * (ell / ell_0) ** alpha


class CorrelatedPowerLaws(PowerLaw):
    """As PowerLaw, but requires only the diagonal of the component-component
    dimensions and the correlation coefficient between TWO PL
    """

    def eval(self, ell=None, alpha=None, ell_0=None, amp=None, rho=None):
        """
        Parameters
        ----------
        ell: float or array
            Multipole
        amp: array
            amplitude of the auto-spectra. Shape must be ``(..., 2)``, which
            means that for time being we support only two correlated components
        alpha: float or array
            Spectral index. Shape must be broadcast-compatible with `amp`
        ell_0: float
            Reference ell

        Returns
        -------
        cl: ndarray
            The dimensions are ``(..., comp1, comp2, ell)``.
            The leading dimensions are the hypothetic dimensions of `alpha` and
            `amp`.
        """
        alpha = np.array(alpha)
        amp = np.array(amp)

        alpha = (alpha[..., np.newaxis, :] + alpha[..., np.newaxis]) / 2
        amp = (amp[..., np.newaxis, :] * amp[..., np.newaxis]) ** 0.5
        amp[..., 1, 0] *= rho
        amp[..., 0, 1] *= rho

        return super().eval(ell=ell, alpha=alpha, ell_0=ell_0, amp=amp)


class PowerSpectraAndCorrelation(Model):
    r"""Components' spectra and their correlation

    Spectrum of correlated components defined by the spectrum of each component
    and their correlation

    Parameters
    ----------
    *power_spectra : series of `PowerSpectrum`
        The series has lenght :math:`N (N + 1) / 2`, where :math:`N` is the
        number of components. They specify the upper (or lower) triangle of the
        component-component cross spectra, which is symmetric. The series stores
        a `PowerSpectrum` for each component-component combination.
        The main diagonal (i.e. the autospectra) goes first, the second diagonal
        of the correlation matrix follows, then the third, etc.
        The ordering is similar to the one returned by `healpy.anafast`.
    """

    def __init__(self, *power_spectra, **kwargs):
        self._power_spectra = power_spectra
        self.n_comp = np.rint(-1 + np.sqrt(1 + 8 * len(power_spectra))) // 2
        self.n_comp = int(self.n_comp)
        assert (self.n_comp + 1) * self.n_comp // 2 == len(power_spectra)
        self.set_defaults(**kwargs)

    def set_defaults(self, **kwargs):
        if "kwseq" in kwargs:
            for i, cl_kwargs in enumerate(kwargs["kwseq"]):
                self._power_spectra[i].set_defaults(**cl_kwargs)

    @property
    def defaults(self):
        return {"kwseq": [ps.defaults for ps in self.power_spectra]}

    def _get_repr(self):
        return {type(self).__name__: [ps._get_repr() for ps in self.power_spectra]}

    def eval(self, kwseq=None):
        """Compute the SED with the given frequency and parameters.

        Parameters
        ----------
        *argss
            The length of `argss` has to be equal to the number of SEDs joined.
            ``argss[i]`` is the argument list of the ``i``-th SED.

        """
        spectra = np.array(
            [ps(**kwargs) for ps, kwargs in zip(self._power_spectra, kwseq)]
        )
        corrs = spectra[self.n_comp :]
        cls = spectra[: self.n_comp]
        sqrt_cls = [np.sqrt(cl) for cl in cls]
        cls_shape = np.broadcast(*cls).shape

        res = np.empty(  # Shape is (..., comp, comp, ell)
            cls_shape[1:-1] + (self.n_comp, self.n_comp) + cls_shape[-1:]
        )

        for i in range(self.n_comp):
            res[..., i, i, :] = cls[i]

        i_corr = 0
        for k_off_diag in range(1, self.n_comp):
            for el_off_diag in range(self.n_comp - k_off_diag):
                i = el_off_diag
                j = el_off_diag + k_off_diag
                res[..., i, j, :] = sqrt_cls[i] * sqrt_cls[j] * corrs[i_corr]
                res[..., j, i, :] = res[..., i, j, :]
                i_corr += 1

        """
        i_corr = 0
        for i in range(self.n_comp):
            res[..., i, i, :] = cls[i]
            for j in range(i + 1, self.n_comp):
                res[..., i, j, :] = sqrt_cls[i] * sqrt_cls[j] * corrs[i_corr]
                res[..., j, i, :] = res[..., i, j, :]
                i_corr += 1
        """

        assert i_corr == len(corrs)
        return res


class PowerSpectraAndCovariance(Model):
    r"""Components' spectra and their covariance

    Spectrum of correlated components defined by the spectrum of each component
    and their correlation

    Parameters
    ----------
    *power_spectra : series of `PowerSpectrum`
        The series has length :math:`N (N + 1) / 2`, where :math:`N` is the
        number of components. They specify the upper (or lower) triangle of the
        component-component cross spectra, which is symmetric. The series stores
        a `PowerSpectrum` for each component-component combination.
        The main diagonal (i.e. the autospectra) goes first, the second diagonal
        of the covariance matrix follows, then the third, etc.
        The ordering is similar to the one returned by `healpy.anafast`.
    """

    def __init__(self, *power_spectra, **kwargs):
        self._power_spectra = power_spectra
        self.n_comp = np.rint(-1 + np.sqrt(1 + 8 * len(power_spectra))) // 2
        self.n_comp = int(self.n_comp)
        assert (self.n_comp + 1) * self.n_comp // 2 == len(power_spectra)
        self.set_defaults(**kwargs)

    def set_defaults(self, **kwargs):
        if "kwseq" in kwargs:
            for i, cl_kwargs in enumerate(kwargs["kwseq"]):
                self._power_spectra[i].set_defaults(**cl_kwargs)

    @property
    def defaults(self):
        return {"kwseq": [ps.defaults for ps in self._power_spectra]}

    def _get_repr(self):
        return {type(self).__name__: [ps._get_repr() for ps in self._power_spectra]}

    def eval(self, kwseq=None):
        """Compute the Cl with the given frequency and parameters.

        kwseq
            The length of `argss` has to be equal to the number of SEDs joined.
            ``kwseq[i]`` is the argument list of the ``i``-th SED.
        """
        spectra = np.array(
            [ps(**kwargs) for ps, kwargs in zip(self._power_spectra, kwseq)]
        )
        res = np.empty(  # Shape is (..., comp, comp, ell)
            spectra.shape[1:-1] + (self.n_comp, self.n_comp) + spectra.shape[-1:]
        )

        i_corr = 0
        for k_off_diag in range(0, self.n_comp):
            for el_off_diag in range(self.n_comp - k_off_diag):
                i = el_off_diag
                j = el_off_diag + k_off_diag
                res[..., i, j, :] = spectra[i_corr]
                res[..., j, i, :] = res[..., i, j, :]
                i_corr += 1

        assert i_corr == len(spectra)
        return res


class SZxCIB_Reichardt2012(PowerSpectraAndCorrelation):
    """PowerSpectrum for SZxCIB (Dunkley et al. 2013)."""

    def __init__(self, **kwargs):
        """Intialize object with parameters."""
        power_spectra = [
            PowerSpectrumFromFile(_get_power_file("tsz_150_bat")),
            PowerLaw(),
            lambda xi: xi,
        ]
        super().__init__(*power_spectra)


class SZxCIB_Addison2012(PowerSpectraAndCovariance):
    """PowerSpectrum for SZxCIB (Dunkley et al. 2013)."""

    def __init__(self, **kwargs):
        """Intialize object with parameters."""
        power_spectra = [
            PowerSpectrumFromFile(_get_power_file("tsz_150_bat")),
            PowerLaw(),
            PowerSpectrumFromFile(_get_power_file("sz_x_cib")),
        ]
        super().__init__(*power_spectra, **kwargs)


# CMB template cls?



import pytest

from fgspectra import cross as fgc
from fgspectra import power as fgp
from fgspectra import frequency as fgf
import numpy as np
import os


def test_ksz():
    assert fgp.kSZ_bat() is not None


def test_ACT_models():
    # define the models from fgspectra
    ksz = fgc.FactorizedCrossSpectrum(fgf.ConstantSED(), fgp.kSZ_bat())
    cibp = fgc.FactorizedCrossSpectrum(fgf.ModifiedBlackBody(), fgp.PowerLaw())
    radio = fgc.FactorizedCrossSpectrum(fgf.PowerLaw(), fgp.PowerLaw())
    cirrus = fgc.FactorizedCrossSpectrum(fgf.PowerLaw(), fgp.PowerLaw())

    # if there are correlations between components,
    # have to define them in a joined spectrum
    tSZ_and_CIB = fgc.CorrelatedFactorizedCrossSpectrum(
        fgf.Join(fgf.ThermalSZ(), fgf.CIB()), fgp.SZxCIB_Addison2012()
    )

    # for testing purposes we'll also compute the tSZ and clustered CIB alone
    tsz = fgc.FactorizedCrossSpectrum(fgf.ThermalSZ(), fgp.tSZ_150_bat())
    cibc = fgc.FactorizedCrossSpectrum(fgf.CIB(), fgp.PowerLaw())

    template_path = os.path.join(os.path.dirname(os.path.abspath(fgp.__file__)), "data")
    cibc_file = os.path.join(template_path, "cl_cib_Choi2020.dat")
    cibc_Choi2020 = fgc.FactorizedCrossSpectrum(
        fgf.CIB(), fgp.PowerSpectrumFromFile(cibc_file)
    )
    tSZ_and_CIB_Choi2020 = fgc.SZxCIB_Choi2020()

    ells = np.array([2000])

    par = {
        "nu_0": 150.0,
        "ell_0": 3000,
        "T_CMB": 2.725,
        "T_d": 9.7,
        "a_tSZ": 4.66,
        "a_kSZ": 1.60,
        "a_p": 6.87,
        "beta_p": 2.08,
        "a_c": 6.10,
        "beta_c": 2.08,
        "n_CIBC": 1.20,
        "xi": 0.09,
        "a_s": 3.50,
        "a_g": 0.88,
        "f0_sz": 146.9,
        "f0_synch": 147.6,
        "f0_dust": 149.7,
    }

    fsz = np.array([par["f0_sz"]])
    fsynch = np.array([par["f0_synch"]])
    fdust = np.array([par["f0_dust"]])

    result = (
        par["a_tSZ"]
        * tsz({"nu": fsz, "nu_0": par["nu_0"]}, {"ell": ells, "ell_0": par["ell_0"]}),
        par["a_kSZ"] * ksz({"nu": fsz}, {"ell": ells, "ell_0": par["ell_0"]}),
        par["a_p"]
        * cibp(
            {
                "nu": fdust,
                "nu_0": par["nu_0"],
                "temp": par["T_d"],
                "beta": par["beta_p"],
            },
            {"ell": ells, "ell_0": par["ell_0"], "alpha": 2},
        ),
        par["a_c"]
        * cibc(
            {
                "nu": fdust,
                "nu_0": par["nu_0"],
                "temp": par["T_d"],
                "beta": par["beta_c"],
            },
            {"ell": ells, "ell_0": par["ell_0"], "alpha": 2 - par["n_CIBC"]},
        ),
        tSZ_and_CIB(
            {
                "kwseq": (
                    {"nu": fsz, "nu_0": par["nu_0"]},
                    {
                        "nu": fdust,
                        "nu_0": par["nu_0"],
                        "temp": par["T_d"],
                        "beta": par["beta_c"],
                    },
                )
            },
            {
                "kwseq": (
                    {"ell": ells, "ell_0": par["ell_0"], "amp": par["a_tSZ"]},
                    {
                        "ell": ells,
                        "ell_0": par["ell_0"],
                        "alpha": 2 - par["n_CIBC"],
                        "amp": par["a_c"],
                    },
                    {
                        "ell": ells,
                        "ell_0": par["ell_0"],
                        "amp": -par["xi"] * np.sqrt(par["a_tSZ"] * par["a_c"]),
                    },
                )
            },
        ),
        par["a_s"]
        * radio(
            {"nu": fsynch, "nu_0": par["nu_0"], "beta": -0.5 - 2},
            {"ell": ells, "ell_0": par["ell_0"], "alpha": 2},
        ),
        par["a_g"]
        * cirrus(
            {"nu": fdust, "nu_0": par["nu_0"], "beta": 3.8 - 2},
            {"ell": ells, "ell_0": par["ell_0"], "alpha": -0.7},
        ),
    )

    """
    The following array `dunk13` was generated by inserting this code block to
    line 188 of `ACT_equa_likelihood.f90` in the Fortran ACT multifrequency
    likelihood.

    ```fortran
    if(il==2000) then
      write(*,*) (f1*f1)/(f0*f0)*amp_tsz*cl_tsz(il)
      write(*,*) amp_ksz*cl_ksz(il)
      write(*,*) amp_d*cl_p(il)*(f1_dust/feff)**(2.d0*beta_d)*(planckratiod1*fluxtempd1)**2.d0
      write(*,*) amp_c*cl_c(il)*(f1_dust/feff)**(2.d0*beta_c)*(planckratiod1*fluxtempd1)**2.d0
      write(*,*) -2.d0*sqrt(amp_c*amp_tsz)*xi*cl_szcib(il)*((f1_dust**beta_c*f1*planckratiod1*fluxtempd1)/(feff**beta_c*f0))
      write(*,*) amp_s*cl_p(il)*(f1_synch/feff)**(2.d0*alpha_s)*fluxtemps1**2.d0
      write(*,*) amp_ge*cl_cir(il)*(f1_dust**2.d0/feff**2.d0)**beta_g*fluxtempd1**2.d0
    endif
    ```

	The array 'choi20' was generated by inserting this code block to 
    line 241 of `ACTPol_deep.f90` in the Fortran ACT multifrequency
    likelihood and by changing the default value of 'fsz', 'fdust', 
    and 'fsyn' to match the one used in this test.

    ```fortran
    if(il==2000) then
    	write(*,*) a_c*clc(il)*(fdust(2)/feff)**(2.d0*beta_c)*(planckratiod(2)*fluxtempd(2))**2.d0
    	write(*,*) -2.d0*sqrt(a_c*a_tsz)*xi*clszcib(il)*((fdust(2)**beta_c*fszcorr(2)*planckratiod(2)*fluxtempd(2))/(feff**beta_c*f0))
    endif
	```
    """
    dunk13 = np.array(
        [
            4.46395567509345,  # tsz
            1.37903889069153,  # ksz
            3.02039291213271,  # cibp
            4.36478789541715,  # cibc
            -0.611930900267790,  # cibxtsz
            1.63099887443075,  # sync
            1.15557760337592,
        ]
    )  # cirrus

    # for testing purposes, subtract tsz and cibc alone
    # from the combined tsz,cibc,tszxcibc.
    result = np.array([r[0][0][0] for r in result])
    result[-3] -= result[0]
    result[-3] -= result[3]

    tsz_fg, ksz_fg, cibp_fg, cibc_fg, cibxtxz_fg, sync_fg, cirrus_fg = result

    assert np.abs(tsz_fg - dunk13[0]) / dunk13[0] < 1e-3
    assert np.abs(ksz_fg - dunk13[1]) / dunk13[1] < 1e-3
    assert np.abs(cibp_fg - dunk13[2]) / dunk13[2] < 1e-3
    assert np.abs(cibc_fg - dunk13[3]) / dunk13[3] < 1e-3
    assert np.abs(cibxtxz_fg - dunk13[4]) / dunk13[4] < 1e-3
    assert np.abs(sync_fg - dunk13[5]) / dunk13[5] < 1e-3
    assert np.abs(cirrus_fg - dunk13[6]) / dunk13[6] < 1e-3
